/* Copyright (c) 2014-2016, Marel
 * Copyright (c) 2023, Andri Yngvason
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#pragma once

#define HTTP_FIELD_INDEX_MAX 32

#include <stddef.h>

struct http_kv {
	char* key;
	char* value;
};

struct http_req {
	size_t header_length;
	size_t content_length;
	char* content_type;
	size_t field_index;
	struct http_kv field[HTTP_FIELD_INDEX_MAX];
};

int http_req_parse(struct http_req* req, const char* head);
void http_req_free(struct http_req* req);
